import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time
#import multiprocessing as mp
from curvedpy import Conversions

# -------------------------------
# ----NAMING CONVENTIONS USED----
# -------------------------------
#
# UPPER & LOWER INDICES
# _ indicates an upper/contravariant index
# __ indicates a lower/covariant index
# For example g__mu__nu is the metric g with two lower indices while g_mu_nu is the metric 
# with two upper indices, which is the inverse of g__mu__nu
# Another example is 4-momentum k_mu, which has a 0 (time) component of k_0 or k_t
# The 4-momentum as a oneform is then k__mu and its zeroth component is k__0
# An example tensor could be T_mu__nu__alp. This tensor has one upper/contravariant (mu) and 
# two lower/covariant/oneform (nu, alp) indices.
#
# COORDINATES
# the used coordinate system for a tensor is indicated with by appending for example 
# _xyz, _sph, _bl after the indices.
# Example: x_mu_bl: this 4vector with one upper index is given in Boyer-Lindquist coordinates
# Example: g__mu__nu_sph: this metric tensor with two lower indices is given in spherical coordinates
#
# MISC
# If a vector x_mu has only 3 components, they are the three spatial components



class GeodesicIntegratorSchwarzschildXYZ:

    conversions = Conversions()

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, time_like = False, verbose=False):



        self.M = mass
        self.r_s_value = 2*self.M 

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet
        self.verbose = verbose
        if verbose:
            print("Geodesic SS Integrator Settings: ")
            print(f"  - {self.M=}")
            print(f"  - {self.r_s_value=}")
            print(f"  - {self.time_like=}")
            print(f"  - {self.verbose=}")
            print("--")

        # Define symbolic variables
        self.t, self.r, self.th, self.ph, self.r_s = sp.symbols("t r \\theta \\phi r_s", real=True)
        self.x, self.y, self.z = sp.symbols('x, y, z', real=True)

        self.g__mu__nu_sph = sp.Matrix([\
            [-1*(1-self.r_s/self.r), 0, 0, 0],\
            [0, 1/(1-self.r_s/self.r), 0, 0],\
            [0, 0, self.r**2, 0],\
            [0, 0, 0, self.r**2 * sp.sin(self.th)**2]\
            ])

        self.g__mu__nu_sph_lamb = sp.lambdify([self.t, self.r, self.th, self.ph, self.r_s], self.g__mu__nu_sph, "numpy")
        
        # self.to_cart_sub_list = [(self.th, sp.acos(self.z/sp.sqrt(self.x**2+self.y**2+self.z**2))), \
        #                          (self.r,  sp.sqrt(self.x**2+self.y**2+self.z**2)),\
        #                          (self.ph, sp.atan2(self.y, self.x))]
        self.to_cart_sub_list = [(self.th, sp.acos(self.z/(self.x**2+self.y**2+self.z**2)**0.5)), \
                                 (self.r,  (self.x**2+self.y**2+self.z**2)**0.5),\
                                 (self.ph, sp.atan2(self.y, self.x))]
        print("g")
        self.g__mu__nu_cart = self.g__mu__nu_sph.subs(self.to_cart_sub_list)
        self.g__mu__nu_cart.simplify()
        print("g_inv")
        self.g_mu_nu_cart = self.g__mu__nu_cart.inv()
        #self.g_inv.subs(self.to_cart_sub_list)
        self.g_mu_nu_cart.simplify()

        #YOU ALSO NEED TO TRANSFORM IT!
        print("transforming")
        self.g__mu__nu_cart = self.trans_to_xyz(self.g__mu__nu_cart)
        self.g_mu_nu_cart = self.trans_inv_to_xyz(self.g_mu_nu_cart)

        # I WOULD FIRST DIFF
        print("g_diff")
        self.g__mu__nu_cart_diff = [self.g__mu__nu_cart.diff(self.t), self.g__mu__nu_cart.diff(self.x), \
                                    self.g__mu__nu_cart.diff(self.y), self.g__mu__nu_cart.diff(self.z)]

        # We lambdify these to get numpy arrays
        print("g_lamb")
        self.g__mu__nu_cart_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g__mu__nu_cart)
        self.g_mu_nu_cart_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g_mu_nu_cart)
        self.g__mu__nu_cart_diff_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g__mu__nu_cart_diff)

        # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1)
        self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
        self.k_mu_cart = sp.Matrix([self.k_t, self.k_x, self.k_y, self.k_z])
        self.norm_k = (self.k_mu_cart.T*self.g__mu__nu_cart*self.k_mu_cart)[0]

        self.norm_k_lamb = sp.lambdify([self.k_t, self.k_x, self.k_y, self.k_z, self.x, self.y, self.z, \
                                               self.r_s], self.norm_k, "numpy")


    ################################################################################################
    # 
    ################################################################################################

    def trans_to_xyz(self, g__mu__nu_cart):
        r = sp.sqrt(self.x**2+self.y**2+self.z**2)
        th = sp.acos(self.z/r)
        phi = sp.atan2(self.y,self.x)
        M = sp.Matrix([[1, 0, 0, 0],\
                        [0, r.diff(self.x), r.diff(self.y), r.diff(self.z)],\
                        [0, th.diff(self.x), th.diff(self.y), th.diff(self.z)],\
                        [0, phi.diff(self.x), phi.diff(self.y), phi.diff(self.z)],\
                     ])

        return M.T*g__mu__nu_cart*M # !check


    def trans_inv_to_xyz(self, g_mu_nu):
        x = self.r * sp.sin(self.th) * sp.cos(self.ph)
        y = self.r * sp.sin(self.th) * sp.sin(self.ph)
        z = self.r * sp.cos(self.th)
        M = sp.Matrix([[1, 0, 0, 0],\
                      [0, x.diff(self.r), x.diff(self.th), x.diff(self.ph)],\
                      [0, y.diff(self.r), y.diff(self.th), y.diff(self.ph)],\
                      [0, z.diff(self.r), z.diff(self.th), z.diff(self.ph)],\
                     ]).subs(self.to_cart_sub_list)

        return M*g_mu_nu*M.T 


    ################################################################################################
    #
    ################################################################################################
    def get_dk(self, kt_val, kx_val, ky_val, kz_val, t_val, x_val, y_val, z_val):
        # Calc g, g_inv and g_diff at given coords
        g = self.g__mu__nu_cart_lamb(t_val, x_val, y_val, z_val, self.r_s_value)
        g_inv = self.g_mu_nu_cart_lamb(t_val, x_val, y_val, z_val, self.r_s_value)
        g_diff = self.g__mu__nu_cart_diff_lamb(t_val, x_val, y_val, z_val, self.r_s_value)

        # Calc the connection Symbols at given coords
        gam_t = sp.Matrix([[self.gamma_func(g, g_inv, g_diff, 0, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_x = sp.Matrix([[self.gamma_func(g, g_inv, g_diff, 1, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_y = sp.Matrix([[self.gamma_func(g, g_inv, g_diff, 2, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_z = sp.Matrix([[self.gamma_func(g, g_inv, g_diff, 3, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])

        # Building up the geodesic equation: 
        # Derivatives: k_beta = d x^beta / d lambda
        k = [kt_val, kx_val, ky_val, kz_val]
    
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        dk_t = sum([- gam_t[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        dk_x = sum([- gam_x[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        dk_y = sum([- gam_y[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        dk_z = sum([- gam_z[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])

        return dk_t, dk_x, dk_y, dk_z

    # Connection Symbols
    def gamma_func(self, g, g_inv, g_diff, sigma, mu, nu):

        g_sigma_mu_nu = 0
        for rho in [0,1,2,3]:
            if g[sigma, rho] != 0:
                g_sigma_mu_nu += 1/2 * g_inv[sigma, rho] * (\
                                g_diff[mu][nu, rho] + \
                                g_diff[nu][rho, mu] - \
                                g_diff[rho][mu, nu] )
            else:
                g_sigma_mu_nu += 0
        return g_sigma_mu_nu

    def k_t_from_norm(self, k0, x0, t=0):
        sub_list = [(self.k_x, k0[0]),\
                    (self.k_y, k0[1]),\
                    (self.k_z, k0[2]),\
                    (self.t, t),\
                    (self.x, x0[0]),\
                    (self.y, x0[1]),\
                    (self.z, x0[2]),\
                    (self.r_s, self.r_s_value),\
                    ]
        norm_k_subbed = self.norm_k.subs(sub_list)

        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle
        # time_like = False: calculates a geodesic for a photon
        if (self.time_like):
            k_t_from_norm = sp.solve(norm_k_subbed+1, self.k_t)#[1]
        else:
            k_t_from_norm = sp.solve(norm_k_subbed, self.k_t)#[1]
        if len(k_t_from_norm) > 1:
            k_t_from_norm = k_t_from_norm[1]
        return float(k_t_from_norm)
        # self.k_t_from_norm_lamb = sp.lambdify([self.k_x, self.k_y, self.k_z, self.x, self.y, self.z, self.t, self.r_s], \
        #                                         self.k_t_from_norm, "numpy")

    def calc_oneform(self, v_mu, g__mu__nu):
        v_nu = g__mu__nu * v_mu
        return v_nu


    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory(self, k0_xyz, x0_xyz, *args, **kargs):
        #mp_on = False

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)


        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if len(k0_xyz) == 1:
            return self.calc_trajectory_xyz(k0_xyz[0], x0_xyz[0], *args, **kargs)
        else:
            return [self.calc_trajectory_xyz(k0_xyz[i], x0_xyz[i], *args, **kargs) for i in range(len(x0_xyz))]

    ################################################################################################
    # calc_trajectory
    ################################################################################################
    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def calc_trajectory_xyz(self, \
                        k0_xyz = np.array([1, 0.0, 0.0]), x0_xyz = np.array([-10, 10, 0]), \
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        # Calculate from norm of starting condition
        k0_t = self.k_t_from_norm(k0=k0_xyz, x0= x0_xyz, t = 0)#, r_s = self.r_s_value)
        #k0_t = self.k_t_from_norm_lamb(*k0_xyz, *x0_xyz, t = 0, r_s = self.r_s_value)
        values_0 = [k0_t, *k0_xyz, *x0_xyz]#[ k0_x, x0, k0_y, y0, k0_z, z0, k0_t ]


        r0 = np.linalg.norm(x0_xyz)
        if R_end == -1:
            R_end = np.inf
        elif R_end < r0:
            R_end = np.sqrt(x0**2 + y0**2 + z0**2)*1.01

        if r0 > self.r_s_value:
            # Step function needed for solve_ivp
            def step(lamb, new):
                k_t_new, k_x_new, k_y_new, k_z_new, x_new, y_new, z_new = new

                #print(new)
                
                dk_t, dk_x, dk_y, dk_z = self.get_dk(k_t_new, k_x_new, k_y_new, k_z_new, \
                                                    t_val=lamb, x_val=x_new, y_val=y_new, z_val=z_new)

                dx, dy, dz = k_x_new, k_y_new, k_z_new

                #new_k_x, new_x, new_k_y, new_y, new_k_z, new_z, new_k_t = new

                # new_dk_t = self.dk_t_lamb(*new, t = lamb, r_s = self.r_s_value)
                # #dr = new_k_r
                # new_dk_x = self.dk_x_lamb(*new, t = lamb, r_s = self.r_s_value)
                # dx = new_k_x
                # new_dk_y = self.dk_y_lamb(*new, t = lamb, r_s = self.r_s_value)
                # dy = new_k_y
                # new_dk_z = self.dk_z_lamb(*new, t = lamb, r_s = self.r_s_value)
                # dz = new_k_z

                # return( new_dk_x, dx, new_dk_y, dy, new_dk_z, dz, new_dk_t)
                return( dk_t, dk_x, dk_y, dk_z, dx, dy, dz)

            def hit_blackhole(t, y): 
                eps = 0.01
                #k_x, x, k_y, y, k_z, z, k_t = y
                k_t, k_x, k_y, k_z, x, y, z = y
                #if verbose: print("Event - hit_blackhole: ", r-self.r_s_value)
                return np.sqrt(x**2+y**2+z**2) - (self.r_s_value+eps)
            hit_blackhole.terminal = True

            def reached_end(t, y): 
                #k_x, x, k_y, y, k_z, z, k_t = y
                k_t, k_x, k_y, k_z, x, y, z = y
                return np.sqrt(x**2+y**2+z**2) - R_end
            reached_end.terminal = True
            
            if nr_points_curve == 0:
                t_pts = None
            else:
                t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

            start = time.time()
            events = [hit_blackhole]
            events.append(reached_end)

            result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                               events=events,\
                               method=method,\
                               max_step = max_step,\
                               first_step = first_step,\
                               atol=atol,\
                               rtol = rtol)
            end = time.time()
            if verbose: print("New: ", result.message, end-start, "sec")

            result.update({"R_end": R_end})
            result.update({"hit_blackhole": len(result.t_events[0])>0})
            result.update({"start_inside_hole": False})
            result.update({"end_check": len(result.t_events[1])>0})

            k_t, k_x, k_y, k_z, x, y, z = result.y
            t = result.t

            k3 = np.array([k_x, k_y, k_z])
            k4 = np.array([k_t, k_x, k_y, k_z])
            x3 = np.array([x, y, z])
            x4 = np.array([t, x, y, z])
            
            result.update({"k4_xyz": k4, "x4_xyz": x4})
            result.update({"k3_xyz": k3, "x3_xyz": x3})

            return k3, x3, result



        else:
            if verbose: print("Starting location inside the blackhole.")
            result = {"start_inside_hole": True}

            return result




import sympy as sp
import numpy as np


#################################################################################
# TO CART
#################################################################################
def make_J_sph1_to_cart():
    r, th, ph = sp.symbols("r \\theta \\phi")
    x = r * sp.sin(th)*sp.cos(ph)
    y = r * sp.sin(th)*sp.sin(ph)
    z = r * sp.cos(th)
    
    matrix_J_r_th_ph = sp.simplify(\
                            sp.Matrix([[x.diff(r), x.diff(th), x.diff(ph)], \
                           [y.diff(r), y.diff(th), y.diff(ph)], \
                           [z.diff(r), z.diff(th), z.diff(ph)]]))
   
    J = sp.lambdify([r, th, ph], matrix_J_r_th_ph, 'numpy')
    C = sp.lambdify([r, th, ph], [x, y, z], 'numpy')
    Det = sp.lambdify([r, th, ph], matrix_J_r_th_ph.det(), 'numpy')

    return J, C, Det, 

J_sph1_to_cart, coords_sph1_to_cart, Det_sph1_to_cart = make_J_sph1_to_cart()


def make_J_sph2_to_cart():
    r, th2, ph2 = sp.symbols("r \\theta2 \\phi2")
    x = r * sp.sin(th2) * sp.cos(ph2)
    y = r * sp.cos(th2)
    z = r * sp.sin(th2) * sp.sin(ph2)

    matrix_J_r_th2_ph2 = sp.simplify(
                            sp.Matrix([[x.diff(r), x.diff(th2), x.diff(ph2)], \
                           [y.diff(r), y.diff(th2), y.diff(ph2)], \
                           [z.diff(r), z.diff(th2), z.diff(ph2)]]))
   
    J = sp.lambdify([r, th2, ph2], matrix_J_r_th2_ph2, 'numpy')
    C = sp.lambdify([r, th2, ph2], [x, y, z], 'numpy')
    Det = sp.lambdify([r, th2, ph2], matrix_J_r_th2_ph2.det(), 'numpy')

    return J, C, Det, 

J_sph2_to_cart, coords_sph2_to_cart, Det_sph2_to_cart = make_J_sph2_to_cart()



#################################################################################
# FROM CART
#################################################################################
def make_J_cart_to_sph2():
    x, y, z = sp.symbols("x y z")
    r = sp.sqrt(x**2+y**2+z**2)
    th2 = sp.acos(y/r)
    ph2 = sp.atan2(z,x)
    
    matrix_J_x_y_z = sp.simplify(sp.Matrix([[r.diff(x), r.diff(y), r.diff(z)], \
                               [th2.diff(x), th2.diff(y), th2.diff(z)],\
                               [ph2.diff(x), ph2.diff(y), ph2.diff(z)]]))
    
    J = sp.lambdify([x, y, z], matrix_J_x_y_z, 'numpy')
    C = sp.lambdify([x, y, z], [r, th2, ph2], 'numpy')
    Det = sp.lambdify([x, y, z], matrix_J_x_y_z.det(), 'numpy')

    return J, C, Det

J_cart_to_sph2, coords_cart_to_sph2, Det_cart_to_sph2 = make_J_cart_to_sph2()


def make_J_cart_to_sph1():
    x, y, z = sp.symbols("x y z")
    r = sp.sqrt(x**2+y**2+z**2)
    th = sp.acos(z/r)
    ph = sp.atan2(y, x)
    
    matrix_J_x_y_z = sp.simplify(sp.Matrix([[r.diff(x), r.diff(y), r.diff(z)], \
                               [th.diff(x), th.diff(y), th.diff(z)],\
                               [ph.diff(x), ph.diff(y), ph.diff(z)]]))
    
    J = sp.lambdify([x, y, z], matrix_J_x_y_z, 'numpy')
    C = sp.lambdify([x, y, z], [r, th, ph], 'numpy')
    Det = sp.lambdify([x, y, z], matrix_J_x_y_z.det(), 'numpy')

    return J, C, Det

J_cart_to_sph1, coords_cart_to_sph1, Det_cart_to_sph1 = make_J_cart_to_sph1()



#################################################################################
# Check of determinant to see if coord transformation is defined
#################################################################################

def det_check_sph1_to_sph2(x_sph1):
    x_cart = coords_sph1_to_cart(*x_sph1)
    return not ((Det_sph1_to_cart(*x_sph1) == 0.0) or (Det_cart_to_sph2(*x_cart) == 0.0))

def det_check_sph2_to_sph1(x_sph2):
    x_cart = coords_sph2_to_cart(*x_sph2)
    return not ((Det_sph2_to_cart(*x_sph2) == 0.0) or (Det_cart_to_sph1(*x_cart) == 0.0))




#################################################################################
# BETWEEN SPH1 and SPH2 patches
#################################################################################

def sph1_to_cart(k, x):
    if len(k.shape) == 1:
        return J_sph1_to_cart(*x)@k, coords_sph1_to_cart(*x)
    else:
        k, x = np.column_stack(k), np.column_stack(x)
        k_xyz = np.column_stack(np.array([J_sph1_to_cart(*x[i])@k[i] for i in range(len(x))]))
        x_xyz = np.column_stack(np.array([coords_sph1_to_cart(*x[i]) for i in range(len(x))]))
        return k_xyz, x_xyz

def cart_to_sph2(k, x):
    if len(k.shape) == 1:
        return J_cart_to_sph2(*x)@k, coords_cart_to_sph2(*x)
    else:
        k, x = np.column_stack(k), np.column_stack(x)
        k_sph2 = np.column_stack(np.array([J_cart_to_sph2(*x[i])@k[i] for i in range(len(x))]))
        x_sph2 = np.column_stack(np.array([coords_cart_to_sph2(*x[i]) for i in range(len(x))]))
        return k_sph2, x_sph2

def sph1_to_sph2(k, x):
    k_cart, x_cart = sph1_to_cart(k, x)
    k_sph2, x_sph2 = cart_to_sph2(k_cart, x_cart)
    return k_sph2, x_sph2


def sph2_to_cart(k, x):
    if len(k.shape) == 1:
        return J_sph2_to_cart(*x)@k, coords_sph2_to_cart(*x)
    else:
        k, x = np.column_stack(k), np.column_stack(x)
        k_xyz = np.column_stack(np.array([J_sph2_to_cart(*x[i])@k[i] for i in range(len(x))]))
        x_xyz = np.column_stack(np.array([coords_sph2_to_cart(*x[i]) for i in range(len(x))]))
        return k_xyz, x_xyz


def cart_to_sph1(k, x):
    if len(k.shape) == 1:
        return J_cart_to_sph1(*x)@k, coords_cart_to_sph1(*x)
    else:
        k, x = np.column_stack(k), np.column_stack(x)
        k_sph1 = np.column_stack(np.array([J_cart_to_sph1(*x[i])@k[i] for i in range(len(x))]))
        x_sph1 = np.column_stack(np.array([coords_cart_to_sph1(*x[i]) for i in range(len(x))]))
        return k_sph1, x_sph1

def sph2_to_sph1(k, x):
    k_cart, x_cart = sph2_to_cart(k, x)
    k_sph1, x_sph1 = cart_to_sph1(k_cart, x_cart)
    return k_sph1, x_sph1





import sympy as sp
import numpy as np
from scipy.integrate import solve_ivp
import time
import multiprocessing as mp
from curvedpy.utils.conversions import Conversions

class GeodesicIntegratorSchwarzschild_prev_method:

    conversions = Conversions()

    def __init__(self, mass=1.0, time_like = False, verbose=False):

        # Connection Symbols
        def gamma_func(sigma, mu, nu):
            coord_symbols = [self.t, self.r, self.th, self.ph]
            g_sigma_mu_nu = 0
            for rho in [0,1,2,3]:
                if self.g[sigma, rho] != 0:
                    g_sigma_mu_nu += 1/2 * 1/self.g[sigma, rho] * (\
                                    self.g[nu, rho].diff(coord_symbols[mu]) + \
                                    self.g[rho, mu].diff(coord_symbols[nu]) - \
                                    self.g[mu, nu].diff(coord_symbols[rho]) )
                else:
                    g_sigma_mu_nu += 0
            return g_sigma_mu_nu

        self.M = mass
        self.r_s_value = 2*self.M 

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet

        # Define symbolic variables
        self.t, self.r, self.th, self.ph, self.r_s = sp.symbols("t r \\theta \\phi r_s")

        self.g = sp.Matrix([\
            [-1*(1-self.r_s/self.r), 0, 0, 0],\
            [0, 1/(1-self.r_s/self.r), 0, 0],\
            [0, 0, self.r**2, 0],\
            [0, 0, 0, self.r**2 * sp.sin(self.th)**2]\
            ])

        # Connection Symbols
        self.gam_t = sp.Matrix([[gamma_func(0,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_r = sp.Matrix([[gamma_func(1,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_th = sp.Matrix([[gamma_func(2,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_ph = sp.Matrix([[gamma_func(3,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        if verbose: print("Done connection symbols")


        # Building up the geodesic equation: 
        # Derivatives: k_beta = d x^beta / d lambda
        self.k_t, self.k_r, self.k_th, self.k_ph = sp.symbols('k_t k_r k_th k_ph', real=True)
        self.k = [self.k_t, self.k_r, self.k_th, self.k_ph]
    
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        self.dk_t = sum([- self.gam_t[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_r = sum([- self.gam_r[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_th = sum([- self.gam_th[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_ph = sum([- self.gam_ph[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        if verbose: print("Done diff of k")

       # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1)
        self.k = sp.Matrix([self.k_t, self.k_r, self.k_th, self.k_ph])
        self.norm_k = (self.k.T*self.g*self.k)[0]
        self.norm_k_lamb = sp.lambdify([self.k_t, self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                               self.r_s], self.norm_k, "numpy")

        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle (not implemented yet)
        # time_like = False: calculates a geodesic for a photon
        if (self.time_like):
            self.k_t_from_norm = sp.solve(self.norm_k+1, self.k_t)[1]
        else:
            self.k_t_from_norm = sp.solve(self.norm_k, self.k_t)[1]
        if verbose: print("Done norm of k")

        # Lambdify versions
        self.dk_r_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s], \
                                     self.dk_r, "numpy")
        self.dk_th_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s], \
                                     self.dk_th, "numpy")
        self.dk_ph_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                      self.k_t, self.t, self.r_s], \
                                     self.dk_ph, "numpy")
        self.k_t_from_norm_lamb = sp.lambdify([self.k_r, self.r, self.k_th, self.th, self.k_ph, self.ph, \
                                               self.r_s], self.k_t_from_norm, "numpy")
        if verbose: print("Done lambdifying")


    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory(self, k0_xyz, x0_xyz, *args, **kargs):
        mp_on = False

        if not isinstance(k0_xyz, np.ndarray): k0_xyz = np.array(k0_xyz)
        if not isinstance(x0_xyz, np.ndarray): x0_xyz = np.array(x0_xyz)


        if k0_xyz.shape != x0_xyz.shape:
            print("k and x are not the same shape")
            return

        if k0_xyz.ndim == 1:
            if k0_xyz.shape[0] != 3 or x0_xyz.shape[0] != 3:
                print("k or x do not have 3 components")
                return
            k0_xyz = k0_xyz.reshape(1,3)
            x0_xyz = x0_xyz.reshape(1,3)

        else:
            if k0_xyz.shape[1] != 3 or x0_xyz.shape[1] != 3:
                print("k or x do not have 3 components")
                return

        if mp_on == True:
            print("Doing mp", __name__)

            #if __name__ == '__main__':
            print("Got into main")
            pool = mp.Pool(mp.cpu_count())
            print([(k0_xyz_i,x0_xyz_i) for k0_xyz_i, x0_xyz_i in zip(k0_xyz, x0_xyz)])
            results = [pool.apply(self.calc_trajectory_xyz, args=(k0_xyz_i,x0_xyz_i)) for k0_xyz_i, x0_xyz_i in zip(k0_xyz, x0_xyz)]
            pool.close()    

        else:
            return [self.calc_trajectory_xyz(k0_xyz[i], x0_xyz[i], *args, **kargs) for i in range(len(x0_xyz))]


    ################################################################################################
    #
    ################################################################################################
    def calc_trajectory_xyz(self, \
                        k0_xyz = np.array([1, 0.0, 0.0]), x0_xyz = np.array([-10, 10, 0]), \
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        print(curve_start, curve_end)

        if not isinstance(x0_xyz,np.ndarray):
            x0_xyz = np.array(x0_xyz)

        if not isinstance(k0_xyz,np.ndarray):
            k0_xyz = np.array(k0_xyz)

        x0_sph, k0_sph = self.conversions.convert_xyz_to_sph(x0_xyz, k0_xyz)
        k_r_0, k_th_0, k_ph_0 = k0_sph
        r0, th0, ph0 = x0_sph

        result = self.calc_trajectory_sph(\
                        k_r_0 = k_r_0, r0 = r0, k_th_0=k_th_0, th0=th0, k_ph_0=k_ph_0, ph0=ph0,\
                        R_end = R_end, curve_start = curve_start, curve_end = curve_end, nr_points_curve = nr_points_curve, \
                        method = method, max_step = max_step, first_step = first_step, rtol = rtol, atol = atol,\
                        verbose = verbose )
                       

        k_r, r, k_th, th, k_ph, ph = result.y

        k_sph = np.array([k_r, k_th, k_ph])
        x_sph = np.array([r, th, ph])
        x_xyz, k_xyz = self.conversions.convert_sph_to_xyz(x_sph, k_sph)

        return k_xyz, x_xyz, result

    ################################################################################################
    # calc_trajectory_sph
    ################################################################################################
    # This function does the numerical integration of the geodesic equation using scipy's solve_ivp
    def calc_trajectory_sph(self, \
                        k_r_0 = 0., r0 = 10.0, k_th_0 = 0.0, th0 = 1/2*np.pi, k_ph_0 = 0.1, ph0 = 0.0,\
                        R_end = -1,\
                        curve_start = 0, \
                        curve_end = 50, \
                        nr_points_curve = 50, \
                        method = "RK45",\
                        max_step = np.inf,\
                        first_step = None,\
                        rtol = 1e-3,\
                        atol = 1e-6,\
                        verbose = False \
                       ):

        if R_end == -1:
            R_end = np.inf
        elif R_end < r0:
            R_end = r0*1.01

        if r0 > self.r_s_value:
            # Step function needed for solve_ivp
            def step(lamb, new):
                new_k_r, new_r, new_k_th, new_th, new_k_ph, new_ph = new
                new_k_t = self.k_t_from_norm_lamb(*new, self.r_s_value)

                new_dk_r = self.dk_r_lamb(*new, new_k_t, t = 0, r_s = self.r_s_value)
                dr = new_k_r
                new_dk_th = self.dk_th_lamb(*new, new_k_t, t = 0, r_s = self.r_s_value)
                dth = new_k_th
                new_dk_ph = self.dk_ph_lamb(*new, new_k_t, t = 0, r_s = self.r_s_value)
                dph = new_k_ph

                return( new_dk_r, dr, new_dk_th, dth, new_dk_ph, dph)

            def hit_blackhole(t, y): 
                eps = 0.5
                k_r, r, k_th, th, k_ph, ph = y
                #if verbose: print("Test Event Hit BH: ", x, y, z, self.r_s_value, x**2 + y**2 + z**2 - self.r_s_value**2)
                return r - self.r_s_value
            hit_blackhole.terminal = True

            def reached_end(t, y): 
                #k_x, x, k_y, y, k_z, z = y
                k_r, r, k_th, th, k_ph, ph = y
                #print("integrator check end", r, r0)
                #if verbose: print("Test Event End: ", np.sqrt(x**2 + y**2 + z**2), R_end, x**2 + y**2 + z**2 - R_end**2)
                return r - R_end
            reached_end.terminal = True
            
            values_0 = [ k_r_0, r0, k_th_0, th0, k_ph_0, ph0 ]
            if nr_points_curve == 0:
                t_pts = None
            else:
                t_pts = np.linspace(curve_start, curve_end, nr_points_curve)

            start = time.time()
            events = [hit_blackhole]
            events.append(reached_end)

            result = solve_ivp(step, (curve_start, curve_end), values_0, t_eval=t_pts, \
                               events=events,\
                               method=method,\
                               max_step = max_step,\
                               first_step = first_step,\
                               atol=atol,\
                               rtol = rtol)
            end = time.time()
            if verbose: print("New: ", result.message, end-start, "sec")

            result.update({"R_end": R_end})
            result.update({"hit_blackhole": len(result.t_events[0])>0})
            result.update({"start_inside_hole": False})
            result.update({"end_check": len(result.t_events[1])>0})


        else:
            if verbose: print("Starting location inside the blackhole.")
            result = {"start_inside_hole": True}

        return result



# https://betterscientificsoftware.github.io/python-for-hpc/tutorials/python-pypi-packaging/
# https://packaging.python.org/en/latest/tutorials/packaging-projects/

from setuptools import setup, find_packages

VERSION = '0.0.4a0'
DESCRIPTION = 'General Relativistic Geodesic Integrator'
LONG_DESCRIPTION = 'Curvedpy can be used to calculate geodesics (trajectories) for different metrics and for both photons and massive particles.'

# Setting up
setup(
       # the name must match the folder name 'verysimplemodule'
        name="curvedpy", 
        version=VERSION,
        author="B.L. de Vries",
        author_email="<bldevries@protonmail.com>",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=["numpy>=2.0.2", "scipy>=1.13.1", "sympy>=1.13.3", "multiprocess>=0.70.17"], # add any additional packages that 
        license = " GPL-3.0",
        # needs to be installed along with your package. Eg: 'caer'
        url= "https://github.com/bldevries/curvedpy",
        keywords=['python', 'relativistic', 'ray', 'tracer', 'blackhole', 'astronomy', 'physics', 'numerical'],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Education",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ]
)


import numpy as np
import time
from curvedpy.geodesics.blackhole_integrators.schwarzschild_XYZ import GeodesicIntegratorSchwarzschildXYZ
from curvedpy.geodesics.blackhole_integrators.kerrschild_XYZ import GeodesicIntegratorKerrSchildXYZ
from curvedpy.geodesics.blackhole_integrators.schwarzschild_SPH1SPH2 import GeodesicIntegratorSchwarzschildSPH1SPH2


class BlackholeGeodesicIntegrator:
    """
    A class to represent different black hole geodesic integrators as one object.
    
    ...

    Attributes
    ----------
    gi : multiple
        geodesic integrator used

    Methods
    -------
    geodesic(k0_xyz, x0_xyz, *args, **kargs)
        Calculates (a) geodesic(s) and returns its/their trajectories and momentum (per mass)
    """

    # FORCE_COORDINATES_CARTESIAN = {"force_coordinates": "xyz", "Explanation": ""}
    # FORCE_COORDINATES_SPH2PATCH = {"force_coordinates": "SPH2PATCH", "Explanation": ""}
    # FORCE_COORDINATES_SPH = {"force_coordinates": "SPH", "Explanation": "", "Note": "Only for debugging!"}

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, a = 0, time_like = False, coordinates = "SPH2PATCH", theta_switch = 0.1*np.pi, verbose=False):
        """
        Initialize a class to represent different black hole geodesic integrators as one object.

        Parameters
        ----------
            mass : float
                Mass of the black hole in geometrized units
            a : float
                Rotation of the black hole per mass
            time_like: bool
                True for a time-like (massive particles) geodesic and False for a massless (photon) geodesic 
            coordinates : str
                Coordinates used for integration: "xyz" or "SPH2PATCH"
            theta_switch: float
                ...
            verbose: bool
                print or not to print lots of information.

        """

        if a == 0:
            if coordinates == "xyz":
                if verbose: print("Running Schwarzschild integrator in XYZ coords")
                self.gi = GeodesicIntegratorSchwarzschildXYZ(mass=mass, time_like=time_like, verbose = verbose)
            elif coordinates == "SPH2PATCH":
                if verbose: print("Running Schwarzschild integrator in SPH1SPH2 coords")
                self.gi = GeodesicIntegratorSchwarzschildSPH1SPH2(mass=mass, theta_switch = theta_switch, time_like=time_like, verbose = verbose)
            elif coordinates == "SPH": 
                self.gi = GeodesicIntegratorSchwarzschildSPH1SPH2(mass=mass, theta_switch = theta_switch, time_like=time_like, verbose = verbose)
            else:
                print("NO INTEGRATOR SELECTED")

        else:
            if verbose: print("Running KerrSchild integrator in XYZ coords")
            self.gi = GeodesicIntegratorKerrSchildXYZ(mass=mass, a=a, time_like=time_like, verbose = verbose)


    ################################################################################################
    #
    ################################################################################################
    def geodesic(self, k0_xyz, x0_xyz, *args, **kargs):
        """ Calculate a geodesic and return the coordinates and momenta (per mass).

            We always use geometrized units. Time component of the momenta is calculated using the norm of the 4 vector, 
            which is based on the time_like setting. The time component of the initial location is set to 0.

            Keyword arguments:
            k0_xyz -- initial condition, x, y and z component of the 4-momenta (per mass), numpy array of length 3. 
            x0_xyz -- initial condition, x, y and z component of the 4-location, numpy array of length 3
        """

        return self.gi.calc_trajectory(k0_xyz, x0_xyz, *args, **kargs)


    ################################################################################################
    #
    ################################################################################################
    # def get_r_s(self):
    #     """Return the radius of the black hole horizon."""
        
    #     return self.gi.metric.get_r_s()

    def get_m(self):
        """Return the mass of the black hole."""
        return self.gi.M

    def get_metric(self):
        """Get the metric used for integrating the geodesic."""
        return self.gi.metric





import numpy as np

# # Get the impact parameter and vector
# def getImpactParam(self, loc_hit, dir_hit):
#     # We create a line extending the dir_hit vector
#     line = list(zip(*[loc_hit + dir_hit*l for l in range(20)]))
#     # This line is used to construct the impact_vector
#     impact_vector = loc_hit - loc_hit.dot(dir_hit)*dir_hit
#     # We save the length of the impact_vector. This is called the impact parameter in
#     # scattering problems
#     impact_par = np.linalg.norm(impact_vector)
#     # We normalize the impact vector. This way we get, together with dir_hit, an 
#     # othonormal basis
#     if impact_par != 0:
#         impact_vector_normed = impact_vector/impact_par # !!! Check this, gives errors sometimes
#     else:
#         impact_vector_normed = impact_vector

#     return impact_vector_normed, impact_par


# Get the impact parameter and vector
def getImpactParam(ray_origin, ray_direction):
    impact_vector = ray_origin - ray_origin.dot(ray_direction)*ray_direction
    # We save the length of the impact_vector. This is called the impact parameter in
    # scattering problems
    impact_par = np.linalg.norm(impact_vector)
    # We normalize the impact vector. This way we get, together with dir_hit, an 
    # othonormal basis
    if impact_par != 0:
        impact_vector_normed = impact_vector/impact_par # !!! Check this, gives errors sometimes
    else:
        impact_vector_normed = impact_vector

    return impact_vector_normed, impact_par


import sympy as sp
import numpy as np
from curvedpy.utils.conversions_4D import Conversions4D
from scipy.optimize import fsolve


# !!
# Schwarzschild Metrics implemented using Numpy and Sympy
# !!
# Author: B.L. de Vries

################################################################################################
################################################################################################
class SchwarzschildMetricXYZ:
################################################################################################
################################################################################################

    conversions4D = Conversions4D()

    def __init__(self, mass=1.0, time_like = False, verbose=False):
        
        self.M = mass
        self.r_s_value = 2*self.M 

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet
        self.verbose = verbose

        # Define symbolic variables
        #self.t, self.r, self.th, self.ph, self.r_s = sp.symbols("t r \\theta \\phi r_s", real=True)
        self.t, self.x, self.y, self.z, = sp.symbols('t x y z', real=True)
        self.r, self.r_s  = sp.symbols('r r_s', positive=True, real=True)
        self.alp = sp.symbols('alp')
        

        self.r_sub = (self.x**2 + self.y**2 + self.z**2)**sp.Rational(1,2)
        self.alp_sub = self.r_s/(self.r**2*(-self.r_s+self.r)) # NOTE THE r_s is in here for most entries!

        flip_sign_convention = -1

        g00 = -1*( 1-self.r_s/self.r )
        g01, g02, g03 = 0,0,0
        g10 = 0
        g11 = -1*( -1-self.x**2 * self.alp )
        g12 = flip_sign_convention*( -self.x*self.y*self.alp )
        g13 = flip_sign_convention*( -self.x*self.z*self.alp )
        g20 = 0
        g21 = flip_sign_convention*( -self.x*self.y*self.alp )
        g22 = -1*( -1-self.y**2*self.alp )
        g23 = flip_sign_convention*( -self.y*self.z*self.alp )
        g30 = 0
        g31 = flip_sign_convention*( -self.x*self.z*self.alp )
        g32 = flip_sign_convention*( -self.y*self.z*self.alp )
        g33 = -1*( -1 - self.z**2*self.alp )

        self.g__mu__nu_cart = sp.Matrix([[g00,g01,g02,g03], [g10,g11,g12,g13], [g20,g21,g22,g23], [g30,g31,g32,g33]])
        self.g__mu__nu_cart_pre_sub = self.g__mu__nu_cart

        self.g__mu__nu_cart = self.g__mu__nu_cart.subs(self.alp, self.alp_sub).subs(self.r, self.r_sub)

        g_00 = -1*( 1/(1-self.r_s/self.r) )
        g_01, g_02, g_03 = 0,0,0
        g_10 = 0
        g_11 = -1*(-1+self.r_s  * self.x**2/self.r**3)
        g_12 = flip_sign_convention*self.r_s * self.x * self.y/self.r**3
        g_13 = flip_sign_convention*self.r_s * self.x * self.z/self.r**3
        g_20 = 0
        g_21 = flip_sign_convention* self.r_s * self.x * self.y/self.r**3
        g_22 = -1*( -1 + self.r_s * self.y**2/self.r**3 )
        g_23 = flip_sign_convention*( self.r_s * self.y * self.z/self.r**3 )
        g_30 = 0
        g_31 = flip_sign_convention* ( self.r_s * self.x * self.z/self.r**3 )
        g_32 = flip_sign_convention*( self.r_s * self.y * self.z/self.r**3 )
        g_33 = -1*( -1 + self.r_s * self.z**2/self.r**3 )

        self.g_mu_nu_cart = sp.Matrix([[g_00,g_01,g_02,g_03], [g_10,g_11,g_12,g_13], [g_20,g_21,g_22,g_23], [g_30,g_31,g_32,g_33]]).subs(self.r, self.r_sub)
        self.g_mu_nu_cart = self.g_mu_nu_cart.subs(self.alp, self.alp_sub).subs(self.r, self.r_sub)

        self.g__mu__nu_cart_diff = [self.g__mu__nu_cart.diff(self.t), self.g__mu__nu_cart.diff(self.x), \
                                     self.g__mu__nu_cart.diff(self.y), self.g__mu__nu_cart.diff(self.z)]

        # We lambdify these to get numpy arrays
        self.g__mu__nu_cart_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g__mu__nu_cart, "numpy")
        self.g_mu_nu_cart_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g_mu_nu_cart, "numpy")
        self.g__mu__nu_cart_diff_lamb = sp.lambdify([self.t, self.x, self.y, self.z, self.r_s], self.g__mu__nu_cart_diff, "numpy")

        # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1)
        self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
        self.k_mu_cart = sp.Matrix([self.k_t, self.k_x, self.k_y, self.k_z])
        self.norm_k = (self.k_mu_cart.T*self.g__mu__nu_cart*self.k_mu_cart)[0]

        self.norm_k_lamb = sp.lambdify([self.k_t, self.k_x, self.k_y, self.k_z, self.x, self.y, self.z, \
                                               self.r_s], self.norm_k, "numpy")


    ################################################################################################
    #
    ################################################################################################
    def get_dk(self, kt_val, kx_val, ky_val, kz_val, t_val, x_val, y_val, z_val):
        # Calc g, g_inv and g_diff at given coords
        g = self.g__mu__nu_cart_lamb(t_val, x_val, y_val, z_val, self.r_s_value)
        g_inv = self.g_mu_nu_cart_lamb(t_val, x_val, y_val, z_val, self.r_s_value)
        g_diff = self.g__mu__nu_cart_diff_lamb(t_val, x_val, y_val, z_val, self.r_s_value)

        # Calc the connection Symbols at given coords
        gam_t = np.array([[self.gamma_func(g, g_inv, g_diff, 0, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_x = np.array([[self.gamma_func(g, g_inv, g_diff, 1, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_y = np.array([[self.gamma_func(g, g_inv, g_diff, 2, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        gam_z = np.array([[self.gamma_func(g, g_inv, g_diff, 3, mu, nu) for mu in [0,1,2,3]] for nu in [0,1,2,3]])

        # Building up the geodesic equation: 
        # Derivatives: k_beta = d x^beta / d lambda
        #self.k_t, self.k_x, self.k_y, self.k_z = sp.symbols('k_t k_x k_y k_z', real=True)
        #self.k = [self.k_t, self.k_x, self.k_y, self.k_z]
        k = [kt_val, kx_val, ky_val, kz_val]
    
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        dk_t = np.sum(np.array([- gam_t[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
        dk_x = np.sum(np.array([- gam_x[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
        dk_y = np.sum(np.array([- gam_y[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))
        dk_z = np.sum(np.array([- gam_z[nu, mu]*k[mu]*k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]]))

        return dk_t, dk_x, dk_y, dk_z

    # Connection Symbols
    def gamma_func(self, g, g_inv, g_diff, sigma, mu, nu):
        # 29 Jan 2025
        # A list comprehension does NOT speed the following code up!
        # This something like this is NOT better:
        # g_sigma_mu_nu = np.sum(np.array( [ 1/2 * g_inv[sigma, rho] * (g_diff[mu][nu, rho] + g_diff[nu][rho, mu] - g_diff[rho][mu, nu] ) for rho in [0,1,2,3]] ) )
        # This is because you make a list and then sum, while you do not need to make the list. Normal for loop is fasters that I found so far.
        g_sigma_mu_nu = 0
        for rho in [0,1,2,3]:
            g_sigma_mu_nu += 1/2 * g_inv[sigma, rho] * (\
                            g_diff[mu][nu, rho] + \
                            g_diff[nu][rho, mu] - \
                            g_diff[rho][mu, nu] )
        return g_sigma_mu_nu

    def k_t_from_norm(self, k0, x0, t=0):
        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle
        # time_like = False: calculates a geodesic for a photon
        if (self.time_like):
            def wrap(k_t): return self.norm_k_lamb(k_t, k0[0], k0[1], k0[2], x0[0], x0[1], x0[2], self.r_s_value)+1
            k_t_from_norm = fsolve(wrap, 1.0)
        else:
            k_t_from_norm = root = fsolve(self.norm_k_lamb, 1.0, args = (k0[0], k0[1], k0[2], x0[0], x0[1], x0[2], self.r_s_value) )

        return k_t_from_norm[0]

    def oneform(self, k4_mu, x4_mu):

        if k4_mu.shape[0] == 4:
            k4_mu = np.column_stack(k4_mu)
        if x4_mu.shape[0] == 4:
            x4_mu = np.column_stack(x4_mu)

        k4__mu = np.column_stack(np.array([self.g__mu__nu_cart_lamb(*x4_mu[i],self.r_s_value)@k4_mu[i] for i in range(len(k4_mu))]))

        return k4__mu

################################################################################################
################################################################################################
class SchwarzschildMetricSpherical:
################################################################################################
################################################################################################

    conversions4D = Conversions4D()

    ################################################################################################
    #
    ################################################################################################
    def __init__(self, mass=1.0, time_like = False, eps_theta = 0.00000001, verbose=False):

        self.pole_metric_XYZ = SchwarzschildMetricXYZ(mass, time_like, verbose)

        # Connection Symbols
        def gamma_func(sigma, mu, nu):
            coord_symbols = [self.t, self.r, self.th, self.ph]
            gamma_sigma_mu_nu = 0
            for rho in [0,1,2,3]:
                gamma_sigma_mu_nu += 1/2 * self.g_mu_nu[sigma, rho] * (\
                                self.g__mu__nu_diff[mu][nu, rho] + \
                                self.g__mu__nu_diff[nu][rho, mu] - \
                                self.g__mu__nu_diff[rho][mu, nu] )
            return gamma_sigma_mu_nu

        self.M = mass
        #self.r_s_value = 2*self.M 
        self.r_s = 2*self.M 

        # Type of geodesic
        self.time_like = time_like # No Massive particle geodesics yet
        self.verbose = verbose
        if verbose:
            print("Geodesic SS Integrator Settings: ")
            print(f"  - {self.M=}")
            print(f"  - {self.r_s=}")
            print(f"  - {self.time_like=}")
            print(f"  - {self.verbose=}")
            print("--")

        # Define symbolic variables
        self.t, self.r, self.th, self.ph, = sp.symbols('t r \\theta \\phi', real=True)
        #self.r_s  = sp.symbols('r_s', positive=True, real=True)

        self.g__mu__nu = sp.Matrix([\
                            [-1*(1-self.r_s/self.r), 0, 0, 0],\
                            [0, 1/(1-self.r_s/self.r), 0, 0],\
                            [0, 0, self.r**2, 0],\
                            [0, 0, 0, self.r**2 * sp.sin(self.th)**2]\
                            ])

        self.g_mu_nu = self.g__mu__nu.inv()
        self.g__mu__nu_diff = [self.g__mu__nu.diff(self.t), self.g__mu__nu.diff(self.r), \
                                     self.g__mu__nu.diff(self.th), self.g__mu__nu.diff(self.ph)]

        # We lambdify these to get numpy arrays
        self.g__mu__nu_lamb = sp.lambdify([self.t, self.r, self.th, self.ph], self.g__mu__nu, "numpy")
        self.g_mu_nu_lamb = sp.lambdify([self.t, self.r, self.th, self.ph], self.g_mu_nu, "numpy")
        self.g__mu__nu_diff_lamb = sp.lambdify([self.t, self.r, self.th, self.ph], self.g__mu__nu_diff, "numpy")

        # Protect against dev by zero
        self.g_mu_nu = self.g_mu_nu.subs(self.th, self.th+eps_theta) 

        # Connection Symbols
        self.gam_t = sp.Matrix([[gamma_func(0,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_r = sp.Matrix([[gamma_func(1,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_th = sp.Matrix([[gamma_func(2,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        self.gam_ph = sp.Matrix([[gamma_func(3,mu, nu).simplify() for mu in [0,1,2,3]] for nu in [0,1,2,3]])
        if verbose: print("Done connection symbols")

        # Building up the geodesic equation: 
        # Derivatives: k_beta = d x^beta / d lambda
        self.k_t, self.k_r, self.k_th, self.k_ph = sp.symbols('k_t k_r k_th k_ph', real=True)
        self.k = [self.k_t, self.k_r, self.k_th, self.k_ph]
    
        # Second derivatives: d k_beta = d^2 x^beta / d lambda^2
        self.dk_t = sum([- self.gam_t[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_r = sum([- self.gam_r[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_th = sum([- self.gam_th[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        self.dk_ph = sum([- self.gam_ph[nu, mu]*self.k[mu]*self.k[nu] for mu in [0,1,2,3] for nu in [0,1,2,3]])
        if verbose: print("Done diff of k")

       # Norm of k
        # the norm of k determines if you have a massive particle (-1), a mass-less photon (0) 
        # or a space-like curve (1)
        self.k = sp.Matrix([self.k_t, self.k_r, self.k_th, self.k_ph])
        self.norm_k = (self.k.T*self.g__mu__nu*self.k)[0]
        self.norm_k_lamb = sp.lambdify([self.k_t, self.k_r, self.k_th, self.k_ph, self.r, self.th, self.ph], \
                                               self.norm_k, "numpy")

        # Now we calculate k_t using the norm. This eliminates one of the differential equations.
        # time_like = True: calculates a geodesic for a massive particle (not implemented yet)
        # time_like = False: calculates a geodesic for a photon
        if (self.time_like):
            self.k_t_from_norm = sp.solve(self.norm_k+1, self.k_t)[1]
        else:
            self.k_t_from_norm = sp.solve(self.norm_k, self.k_t)[1]
        if verbose: print("Done norm of k")

        # Lambdify versions
        self.dk_t_lamb = sp.lambdify([  self.k_t, self.k_r, self.k_th, self.k_ph, \
                                        self.t, self.r, self.th, self.ph], \
                                        self.dk_t, "numpy")
        self.dk_r_lamb = sp.lambdify([  self.k_t, self.k_r, self.k_th, self.k_ph, \
                                        self.t, self.r, self.th, self.ph], \
                                        self.dk_r, "numpy")
        self.dk_th_lamb = sp.lambdify([ self.k_t, self.k_r, self.k_th, self.k_ph, \
                                        self.t, self.r, self.th, self.ph], \
                                        self.dk_th, "numpy")
        self.dk_ph_lamb = sp.lambdify([ self.k_t, self.k_r, self.k_th, self.k_ph, \
                                        self.t, self.r, self.th, self.ph], \
                                        self.dk_ph, "numpy")
        self.k_t_from_norm_lamb = sp.lambdify([ self.k_r, self.k_th, self.k_ph, \
                                                self.t, self.r, self.th, self.ph], \
                                                self.k_t_from_norm, "numpy")
        if verbose: print("Done lambdifying")

    ################################################################################################
    #
    ################################################################################################
    def get_dk(self, kt_val, kr_val, kth_val, kph_val, t_val, r_val, th_val, ph_val):
        return \
            self.dk_t_lamb(kt_val, kr_val, kth_val, kph_val, t_val, r_val, th_val, ph_val), \
            self.dk_r_lamb(kt_val, kr_val, kth_val, kph_val, t_val, r_val, th_val, ph_val), \
            self.dk_th_lamb(kt_val, kr_val, kth_val, kph_val, t_val, r_val, th_val, ph_val), \
            self.dk_ph_lamb(kt_val, kr_val, kth_val, kph_val, t_val, r_val, th_val, ph_val), \
    
    ################################################################################################
    #
    ################################################################################################
    def get_r_s(self):
        return self.r_s

    ################################################################################################
    #
    ################################################################################################
    def oneform(self, k4_mu, x4_mu):

        if k4_mu.shape[0] == 4:
            k4_mu = np.column_stack(k4_mu)
        if x4_mu.shape[0] == 4:
            x4_mu = np.column_stack(x4_mu)

        k4__mu = np.column_stack(np.array([self.g__mu__nu_lamb(*x4_mu[i])@k4_mu[i] for i in range(len(k4_mu))]))

        return k4__mu


